/* ------------------------------------------------------------------------

	digiDload.c 
	Copyright (C) 1996  Digi International.
 
	For technical support please email digiLnux@dgii.com or
	call Digi tech support at (612) 943-0577

	This code represents our first attempt at providing a universal
	code downloader for our products.  As we decide to support upcoming
	products, this code will be modified to provide that support.  It is
	our hope that much of the card specifics can be abstracted out of
	the kernel and into this application.  This will allow for easier
	development as well as debug, and enable longer release schedules
	on the driver itself (Since it will have to be changed less often).  

 	This program is free software; you can redistribute it and/or modify
 	it under the terms of the GNU General Public License as published by
 	the Free Software Foundation; either version 2 of the License, or
 	(at your option) any later version.

 	This program is distributed in the hope that it will be useful,
 	but WITHOUT ANY WARRANTY; without even the implied warranty of
 	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 	GNU General Public License for more details.

 	You should have received a copy of the GNU General Public License
 	along with this program; if not, write to the Free Software
 	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

--------------------------------------------------------------------------- */


#include <stdio.h>
#include <fcntl.h>
#include <errno.h>
#include <unistd.h>
#include <stdarg.h>
#include <ctype.h>
#include <string.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/mman.h>
#include <stdio.h>
#include <stdlib.h>

/*  ----------------------------------------------------------------
	NOTE: In asm/io.h outb seems to have its aruments reversed
	from the norm.
------------------------------------------------------------------- */

#include <asm/io.h>
#include <sys/termio.h>
#include <linux/tqueue.h>
#include <sys/ioctl.h>

#include <linux/universal.h>
#include <linux/digi1.h>
#include <linux/epca.h>


#define TRUE  1
#define FALSE 0
#define BIOS 1
#define FEP  2
#define CONC 3

#define MAXCARDS 7

/* There should be a way to interrogate board for this */
#define RAM_SIZE 32768

/* ---------------- Begin function prototypes ------------------ */

void fatal(const char * format, ...);
void nap(int);
void init(int digiFD);
int reset_cards(int);
int binary_is_running(int card, int);
void wait_for_binary(int card, int);
int load_binary(int card, int);
unsigned long int readRAM(int, int, int, int);
void writeRAM(int, long, int, int, int);


/* ---------------- End function prototypes ------------------ */


struct file_map
{
    int board_type;      /* What card does this go on ?      */
    int binary_type;     /* Is it a BIOS, a FEP, or a CONC ? */
    char *fname;         /* Name of binary in filesystem.    */
    unsigned long addr;  /* Where the binary goes.           */
};

struct board_t 
{
	volatile char * ram_addr; 
	volatile char * control_addr; 
	struct digi_info digi_info;
};

struct board_t board[MAXCARDS] ;

/* ----------------------------------------------------
	Note: PCXE and PCXEVE share the same binaries.
	Null filenames mean we are at end of list.
------------------------------------------------------- */

struct file_map file_map[] = 
{
    {PCXEM, BIOS, "sxbios.bin", 0x1000}, 
    {PCXEM, FEP,  "sxfep.bin",  0x1000},
    {EISAXEM, BIOS, "sxbios.bin", 0x1000}, 
    {EISAXEM, FEP,  "sxfep.bin",  0x1000},
    {PCIXR,  BIOS, "sxbios.bin", 0x1000}, 
    {PCIXR,  FEP,  "sxfep.bin",  0x1000},
    {PCIXRJ, BIOS, "sxbios.bin", 0x1000}, 
    {PCIXRJ, FEP,  "sxfep.bin",  0x1000},
    {PCIXEM, BIOS, "sxbios.bin", 0x1000}, 
    {PCIXEM, FEP,  "sxfep.bin",  0x1000},
    {PCXE,   BIOS, "xxbios.bin", 0x1800}, 
    {PCXE,   FEP,  "xxfep.bin",  0x0},
    {PCXEVE, BIOS, "xxbios.bin", 0x1800}, 
    {PCXEVE, FEP,  "xxfep.bin",  0x0},
    {PCXI,   BIOS,  "xxbios.bin",  0xFF800},
    {PCXI,   FEP,   "xxfep.bin",  0x2000},
    {PC64XE, BIOS,  "xxbios.bin",  0xFF800},
    {PC64XE, FEP,   "xxfep.bin",  0x2000},
    {-1, -1, NULL, -1}		/* End of array */
};

/*  ---------------------------------------------------------------------
	The last argument of each file_map above indicates the address  
	at which the binary is to be loaded.  In case of the PCXE & PCXEVE
	BIOS this should be calculated because it is dependant on the size of
	the binary.	
------------------------------------------------------------------------- */


int verbose = FALSE; /* verbose defaults to true */
int board_count = 0;
char *version = "1.1.0";
                    
/* -------------------- Begin fatal --------------------- */
void fatal(const char * format, ...)
{ /* Begin fatal */

	va_list args ; 
	va_start (args, format);
	vfprintf(stderr, format, args);
	va_end(args);
	exit(1); 

} /* End fatal */

/* -------------------- Begin nap ----------------------- */
void nap(int ms)
{ /* Begin nap */

	struct timeval tv ;

	/* Shouldn't this be in the library somewhere? */

	tv.tv_sec = 0 ; 
	tv.tv_usec = ms * 1000 ; 

	select(0,0,0,0,&tv) ;

} /* End nap */

/* -------------------- Begin writeRAM  ---------------------- */
void writeRAM(int target_addr, long value, int size, int card, int bus)
{ /* Begin writeRAM */

	if (bus == PCI_CTL)
	{ /* Begin PCI_CTL */

		if (size == SIZE8)
			*((int_8 *)&board[card].control_addr[target_addr]) = value;
		else 
		if (size == SIZE16)
			*((int_16 *)&board[card].control_addr[target_addr]) = value;
		else
		if (size == SIZE32)
			*((int_32 *)&board[card].control_addr[target_addr]) = value;
		else
			printf("<Error> - writeRAM unknown size\n");

	} /* End PCI_CTL */
	else
	{ /* Begin non-PCI */

		if (size == SIZE8)
			*((int_8 *)&board[card].ram_addr[target_addr]) = value;
		else 
		if (size == SIZE16)
			*((int_16 *)&board[card].ram_addr[target_addr]) = value;
		else
		if (size == SIZE32)
			*((int_32 *)&board[card].ram_addr[target_addr]) = value;
		else
			printf("<Error> - writeRAM unknown size\n");

	} /* End non-PCI */

} /* End writeRAM */  
/* -------------------- Begin readRAM  ---------------------- */
unsigned long int readRAM(int target_addr, int size, int card, int bus)
{ /* Begin readRAM */

	unsigned long int retVal;
	int_8  val8 = 0;
	int_16 val16 = 0;
	int_32 val32 = 0;

	if (bus == PCI_CTL)
	{ /* Begin PCI_CTL */

		if (size == SIZE8)
			val8 = *((int_8 *)&board[card].control_addr[target_addr]);
		else 
		if (size == SIZE16)
			val16 = *((int_16 *)&board[card].control_addr[target_addr]);
		else
		if (size == SIZE32)
			val32 = *((int_32 *)&board[card].control_addr[target_addr]);
		else
			printf("<Error> - writeRAM unknown size\n");

	} /* End PCI_CTL */
	else
	{ /* Begin normal */

		if (size == SIZE8)
			val8 = *((int_8 *)&board[card].ram_addr[target_addr]);
		else 
		if (size == SIZE16)
			val16 = *((int_16 *)&board[card].ram_addr[target_addr]);
		else
		if (size == SIZE32)
			val32 = *((int_32 *)&board[card].ram_addr[target_addr]);
		else
			printf("<Error> - readRAM unknown size\n");
	} /* End normal */

	retVal = val8 | val16 | val32;
	return(retVal);

} /* End readRAM */

/* -------------------- Begin init ---------------------- */
void init(int digiFD)
{ /* Begin init */

	int x; 
	int kmem_fd; 
	unsigned int ram_size;
	short iopl_active = FALSE;


	/*
		This function initializes our internals.  It should *NOT* touch
		any hardware itself.
	*/

	kmem_fd = open("/dev/kmem", O_RDWR);

	if (kmem_fd < 0 )
	{
		fatal("Open of /dev/kmem failed, errno %d\n", errno);
	}

	for (x = 0; x < MAXCARDS; x++)
	{ /* Begin for each board */
		struct digi_info *di = &board[x].digi_info ;
		di->board = x ; 
		if (ioctl(digiFD, DIGI_GETINFO, di) < 0)
		{
			/* Should we fall completely out here (Not just continue) ? */
			continue;
		}
		board_count++;

		/*
			The membase returned is the physical address of the device
			in memory.  Example 0xd0000, 0xd8000 ...
		*/

		/* Make ram size temporarily = 64K.  This is abritrary */
		if (di->type < PC64XE)
			ram_size = RAM_SIZE;
		else
			ram_size = (RAM_SIZE * 2);
			
		board[x].ram_addr = mmap (NULL, ram_size, PROT_READ|PROT_WRITE, 
		      	                   MAP_SHARED|MAP_FILE , kmem_fd, (off_t)di->membase) ;

		if ((int)(board[x].ram_addr) < 0 )
		{
			perror("Mmap failed");
			fatal("Mmap failed on ram addr , errno %d\n", errno);
			
		}

		/* 
			We check the type identifier to see if the board is PCI.  The
			identifiers for non-PCI boards are all smaller than the identifier
			for the PCIXEM.  If the board is non-PCI we map the boards i/o 
			ports.  Since most Digi PCI boards do not have a seperate address
			space (None that this program supports), we must insure that the
			program does not attempt to map i/o for these boards.  The below 
			check does that.
		*/ 
		if (di->type < PCIXEM)
		{ /* Begin the board detected is NOT a PCI board */

			/* Crank open I/O addresses */

			/* 
				iopl gives us access to all possible 65536 ports.  It would
				be better to just open access to the ports we needed, but the
				closest routine that does that (ioperm) does not handle ports
				larger than 0x3ff.  Since we also support EISA this will not
				work.
			*/

			if (!iopl_active)
			{
				if (iopl(3))
					perror("iopl failed ");
				iopl_active = TRUE;
			}

		} /* End the board detected is NOT a PCI board */
		else
		{ /* Begin the board detected IS a PCI board */
			
			/* 
				Most Digi PCI cards have their control registers mapped
				into memory as opposed to i/o space.  We map those registers
				in here.
			*/

			board[x].control_addr = mmap (NULL, RAM_SIZE, PROT_READ|PROT_WRITE, 
			                       MAP_SHARED|MAP_FILE , kmem_fd, (off_t)di->port) ;


			if (board[x].control_addr < 0 )
			{
				fatal("Mmap failed on PCI control space, errno %d\n", errno);
			}

		} /* End the board detected IS a PCI board */


	} /* End for each board */

	if (board_count == 0)
		fatal("No boards detected.  Exiting now\n");

	/* 
		Turn off the poller.  Bad things happen to those who are squirting 
		in code when the poller comes along.
	*/

	if (ioctl(digiFD, DIGI_POLLER, 1) < 0  )
	{
		fatal("Can't ioctl DIGI_POLLER %d %d\n", digiFD, errno); 
	}

} /* End init */

/* -------------------- Begin reset_cards ------------------- */
int reset_cards(int card)
{ /* Begin reset_cards */

	int index; 
	unsigned long temp;
	struct digi_info *di = &board[card].digi_info ;
	int_16 shiftedAddr;
	int_8  lowAddr, highAddr;

	/* 
		Do whatever it takes to reset the card, and make it available.
		This includes asserting reset, mapping the memory address, and 
		interrupts and possibly setting the windowing mode (Dependant on 
		the card type.
	*/
 
	/*

		For most Digi cards (Including the PCXEM family) the formula for
		calculating the address is as follows :
	
		Set bitfields in t to select interrupt.

		disabled ->  0000h	3 ->  0001h
		       5 ->  0002h	7 ->  0003h
		      10 ->  0004h  11 ->  0005h
		      12 ->  0006h  15 ->  0007h

		Set bitfields 0xFF80 ot according to the bit fields
		0xFF8000h of the base address.

		devb[2] = t and 00FFh
		devb[3] = t / 100h 

		For example, a XEM board with interrupt 5, and base address of 
		0x00D8000h would have the below settings:

		t = 0xD80h + 0002h

		devb[2] = 82h = t and 00FF
		devb[3] = 0Dh = t / 100h

		The PC/XE and PC/XEVE are a little different because you have 
		the option of setting a window.  If no window is set the result
		would be the same (exactly) as the formula above.  However, if
		windows are to be set bit position 0010h should be set in 't' above.

		For example, a XE board with interrupt 5, and base address of 
		0x00D8000h with windows on would have the below settings:

		t = 0xD80h + 0002h + 0010h <- Set windows on

		devb[2] = 92h = t and 00FF
		devb[3] = 0Dh = t / 100h

		Look at the example of the cx configuration on page 17 of the
		fep writers guide for another example.  

	*/


	switch (di->type)
	{ /* Begin switch di->type */

		case PCXEM:
		case EISAXEM:

			outb(0x04, (int)di->port + 0); /* Hold card in a reset */

			for (index = 0; (inb((int)di->port) & 0x0e) != 0x4; index ++) 
			{
				if (index  > 5) 
				{
					if ((di->type) == EISAXEM)
						printf("<Error> - reset_card : EISAXEM not resetting\n");	
					else
						printf("<Error> - reset_card : PCXEM not resetting\n");	

					exit(1);
				}
				nap(200);  /* 1/5th of a second */
			}

			shiftedAddr = ((long)(di->membase)) >> 8;

			lowAddr = shiftedAddr & 0xFF;
			outb(lowAddr, (int)di->port + 2); 

			highAddr = shiftedAddr >> 8;
			outb(highAddr, (int)di->port + 3);

			break;

		case PCIXR:
		case PCIXRJ:
		case PCIXEM:
			writeRAM(0, 0x04, SIZE8, card, PCI_CTL); /* Hold card in reset */
			for (index = 0; ((temp=readRAM(0x0, SIZE8, card, PCI_CTL)) & 0x0e) != 0x4; index ++) 
			{
				if (index  > 5) 
				{
					printf("<Error> - reset_card : PCI not resetting\n");	
					exit(1);
				}
				nap(200);  /* 1/5th of a second */
			}

			break;

		case PCXE:
		case PCXEVE: /* XEVE resets the same as the XE */


			/* 
				The format of the below is a little different
				than the PCXEM case.  The values are calculated
				exactly the same as above, BUT if 8K windows are
				to be used the value 0x10 has to be added to 
				the low byte.  Hence 0x00 (For 0xd0000) becomes
				0x10.  This only occurs if 8K windowing is desired.

				The only options for the new XE (glasgo) is 8*8K
				windows or 2*32K windows. The driver knows nothing
				about 32K windows. 

				Below we automaticly choose windowing.
				We always choose windowing.
			*/

			shiftedAddr = ((long)(di->membase)) >> 8;

			/* We set bit 0x10 to enable windows */
			lowAddr = (shiftedAddr + 0x10) & 0xFF;
			outb(lowAddr, (int)di->port + 2); 

			highAddr = shiftedAddr >> 8;
			outb(highAddr, (int)di->port + 3);

			/* ---------------- Start code to reset board */
			outb(0x06, (int)di->port + 0); /* Hold card in a reset */

			/*
				NOTE : The manual talks about enabling the XE memory.  They would 
				use an 0x6 above and below.  We have been using a 0x4
			*/

			for (index = 0; (inb((int)di->port) & 0x0e) != 0x6; index ++) 
			{
				if (index  > 5) 
				{
					printf("<Error> - reset_card : PCXE not resetting\n");	
					exit(1);
				}
				nap(200);  /* 1/5th of a second */
			}
			break;

		case PC64XE:

			/* 
				The format of the below is a little different
				than the PCXE case.  In particular it is not 
				possible to set the older 64K XE boards (Excluding
				PC/XEVE handled above) to 8K windowing.  The values 
				will be calculated the same as the PCXEM case.
			*/

			shiftedAddr = ((long)(di->membase)) >> 8;

			lowAddr = (shiftedAddr) & 0xFF;
			outb(lowAddr, (int)di->port + 2); 

			highAddr = shiftedAddr >> 8;
			outb(highAddr, (int)di->port + 3);

			/* ---------------- Start code to reset board */
			outb(0x06, (int)di->port + 0); /* Hold card in a reset */

			/*
				NOTE : The manual talks about enabling the XE memory.  They would 
				use an 0x6 above and below.  We have been using a 0x4
			*/

			for (index = 0; (inb((int)di->port) & 0x0e) != 0x6; index ++) 
			{
				if (index  > 5) 
				{
					printf("<Error> - reset_card : PC64XE not resetting\n");	
					exit(1);
				}
				nap(200);  /* 1/5th of a second */
			}
			outb(0x06, (int)di->port + 0); /* Hold card in a reset */

			break;

		case PCXI:

			/*
				Note, the PCXI has its memory options set via. dip 
				switch.  This cannot be changed dynamically hence
				(unlike above) you will not see code here programming
				the memory address or the windowing disposition.
			*/
 
			/* --------- Start code to reset board --------- */
			outb(0x06, (int)di->port + 0); /* Hold card in a reset */


			for (index = 0; (inb((int)di->port) & 0x0e) != 0x6; index ++) 
			{
				if (index  > 5) 
				{
					printf("<Error> - reset_card : PCXI not resetting\n");	
					exit(1);
				}
				nap(200);  /* 1/5th of a second */
			}
			break;

	} /* End switch di->type */
	return (0) ;    

} /* End reset_cards */

/* -------------------- Begin binary_is_running ----------------- */
int binary_is_running(int card, int binary_type)
{ /* Begin binary_is_running */

	struct digi_info *di = &board[card].digi_info ;
	int ret_val = 0 ; 

	/* When necessary, turn on card RAM */

	switch(di->type)
	{
		case PCXEM:
		case EISAXEM:
			outb(0x80 , (int)di->port + 1);
			break;
	}

	switch (binary_type)   
	{ /* Begin switch binary type */

		case FEP:
		{
			char fep_sig0;
			char fep_sig1;

			fep_sig0 = readRAM(0xd20, SIZE8, card, NORMAL);
			fep_sig1 = readRAM(0xd21, SIZE8, card, NORMAL);

			/* 
				If the FEP boots correctly, it should return the ascii
				string "OS" in locations 0xD20 and 0xD21.
			*/

			if ((fep_sig0 == 'O') && (fep_sig1 == 'S') )
				ret_val=1;
			break ; 
		}

		case BIOS:
		{
			char bios_sig0;
			char bios_sig1;

			bios_sig0 = readRAM(0xc00, SIZE8, card, NORMAL);
			bios_sig1 = readRAM(0xc01, SIZE8, card, NORMAL);

			/* 
				If the BIOS boots correctly, it should return the ascii
				string "GD" in locations 0xC00 and 0xC01.
			*/
			if (( bios_sig0 == 'G') && (bios_sig1 == 'D') )
				ret_val=1;
		}
		
	} /* End switch binary type */
	return ret_val ;

} /* End binary_is_running */

/* -------------------- Begin wait_for_binary ----------------- */
void wait_for_binary(int card, int binary_type)
{ /* Begin wait_for_binary */

	int interations ; 

	/* 
		Note when I set this for 30 interations (6 seconds) 
	   it will only boot if there is one concentrator.  Doubling the
      time allows 2 concentrators.  240 is a lot, but it should
	   handle 8 concentrators.
	*/
	for (interations = 0; interations < 480; interations++)
	{
		if (binary_is_running(card, binary_type))
		{
			return ;
		}
		nap(200); /* Expressed in milliseconds == 1/5 second */
	}

	fatal("Could not start %s on card %d\n", 
	binary_type == FEP ? "FEPOS" : "BIOS", card);

} /* End wait_for_binary */

/* -------------------- Begin load_binary -------------------- */
int load_binary(int card, int binary_type)
{ /* Begin load_binary */

	struct file_map *fm = file_map; 
	unsigned int cmdInPtr, target_window;
	unsigned long memory_seg=0, target_addr, target_offset = 0;
	int binaryFD, index, nbytes, ret_val = 0; 
	struct digi_info *di = &board[card].digi_info ;
	char *temp_name, buf[256];

	for (fm = file_map ; fm->binary_type != -1 ; fm++ )
	{ /* Begin loop until we run out of binary files. */

		if (( fm->binary_type == binary_type)  &&
		    ( fm->board_type == di->type ) )
		{
			/* If this board doesn't get this type of binary, bail */
			if ( fm->fname == NULL )
			{
				return(0) ;
			}
			break;
		} 

	} /* End loop until we run out of binary files. */
    
	
	/*
	 	The PCXI and PCXE can have varying amounts of memory onboard.
		Because of this we need to calculate where that memory ends,
		and therefore where we can load bios info.
	*/
	if ((di->type == PCXI) || (di->type == PC64XE))
	{
		/* 
			This memory_seg assumes the XI or XE is a 
			64K board.  We need to add code to query
			this to be sure.  See internal release 110c. 

		*/
		
		memory_seg = 0xf000;  
	}

	target_addr = fm->addr ;
	temp_name = (char *)malloc((size_t)((strlen(fm->fname)) + 11)); 
	*temp_name = 0; /* Make first character null; so strcat can work */
	strcat(temp_name,"/etc/digi/");
	strcat(temp_name,fm->fname);	
	if (verbose)
	{
		printf("digiDload version %s\n",version);
		printf("Downloading %s to %lx\n", temp_name, (long) fm->addr );
	}

	/* 
		Note: programmer can do something stupid here and pass
		a non-buf size alignable value that will make this copy 
		fail.  Don't do that.
	*/
 

	if ((binaryFD = open( temp_name, O_RDONLY )) < 0 ) 
	{
		fatal("Open of %s failed, errno %d\n", fm->fname, errno);
	}

	free((void *)temp_name);


	/* Now, slurp the whole binary in, jam it on the card */

	while ((nbytes = read (binaryFD, buf, sizeof(buf))))
	{ /* Begin while read loop */

		/* 
			This block does whatever is necesary to
			prepare the card to receive this chunk.
		*/

		/* First get all the data on the card; switch windows as necessary */
		switch (di->type)
		{ /* Begin switch di->type */

			case PCIXR:
			case PCIXRJ:
			case PCIXEM:
				switch (binary_type)
				{ /* Begin switch binary_type */

					/* 
						We obviously do not need a switch here, but only
						because this BIOS and FEP loads for this card, are
						done in the same manner and location.  I chose to
						use a switch just to mantain code mechanics consistiency.
					*/

					case BIOS:
					case FEP:

						/* 
							The below will cause the a window switch if window
							is about to be exceeded.
						*/

						/* I do not think we have a 32k window problem anymore */
						target_offset = target_addr;
	
				} /* End switch binary_type */
				break;

			case PCXEM:
			case EISAXEM:
				switch (binary_type)
				{ /* Begin switch binary_type */

					/* 
						We obviously do not need a switch here, but only
						because this BIOS and FEP loads for this card, are
						done in the same manner and location.  I chose to
						use a switch just to mantain code mechanics consistiency.
					*/

					case BIOS:
					case FEP:

						/* 
							The below will cause the a window switch if window
							is about to be exceeded.
						*/

						target_offset = target_addr & (RAM_SIZE - 1);
						target_window = target_addr >> 15; 
	
						outb(0x80 | target_window , (int)di->port + 1);
						break;
				} /* End switch binary_type */
				break;

			case PCXE:
			case PCXEVE: /* Works same as PCXE */
				switch (binary_type)
				{ /* Begin  switch binary_type */

					case BIOS:
						/* 
							The below board[card].ram_addr already points
							at the board mapped in at physical memory.
						*/

						/*
							If target_addr was set to 0x1800 (Assuming 2k bios)
							the same code should NEARLY work.  Watch out for the shift 15
						*/
						/* We mask against 0x2000 because it is an 8K window */
						target_offset = target_addr & (0x2000 - 1);
						/* Later develop code just in case bios > 2K */
						/* FIXME : Do not depend on addr = 0x1800 and window = 0xf */

						outb(0x80 | 0xf , (int)di->port + 1);
						break;

					case FEP:

						/* We mask against 0x2000 because it is an 8K window */
						target_offset = target_addr & (0x2000 - 1);
						/* Select window 1 always */
						outb(0x80 | 0x1 , (int)di->port + 1);

						break;
			} /* End switch binary_type */
			break;

		case PCXI:
		case PC64XE:
			switch (binary_type)
			{ /* Begin  switch binary_type */

				case BIOS:

					outb(0x6 , (int)di->port);
					target_offset = target_addr - (memory_seg << 4);

				break;

				case FEP:

					target_offset = target_addr;
				break;

			} /* End switch binary_type */
			break;

		} /* End switch di->type */


		/* Now, we write the chunk to the card */

		memcpy((void *)&board[card].ram_addr[target_offset], (const void *)buf, nbytes);

		target_addr += nbytes ; 

	} /* End while read loop */

	/*
		Note : When we arrive here the memory window we are currently
		in is NOT guaranteed to be 0.
	*/


	/* Consider puting a check to make sure the code got there okay */
	/* Now, all the code is on the card.  We just have to execute it. */

	switch (di->type)
	{ /* Begin switch di->type */

		case PCIXR:
		case PCIXRJ:
		case PCIXEM:
		{ /* Begin PCI XR, XRJ, XRM case */

			switch(binary_type)
			{ /* End switch binary_type */

				case BIOS:
					/* Set boot vector */
					/* Note - these are long words */

					writeRAM(0, 0x0bf00401, SIZE32, card, NORMAL);
					writeRAM(4, 0x0, SIZE32, card, NORMAL);
					writeRAM(0xc00, 0x0, SIZE16, card, NORMAL);

					break;

				case FEP:

					/* Clear confirm word */
					writeRAM(0xd20, 0x0, SIZE16, card, NORMAL);

					/* Set initial address and BIOS command */
					writeRAM(0xc34, 0xbfc01004L, SIZE32, card, NORMAL);
					writeRAM(0xc30, 0x3L, SIZE32, card, NORMAL);

					break;

			} /* End switch binary_type */

			/* Release reset now */
			writeRAM(0, 0x0, SIZE8, card, PCI_CTL);

			/* Wait for binary (BIOS or FEP) to boot */
			wait_for_binary(card, binary_type);

			break;
		} /* End PCI XR, XRJ, XRM case */

		case PCXEM:
		case EISAXEM:
		{ /* Begin case PCXEM */

			/* Put us back in window 0 */
			outb(0x80, (int)di->port + 1);

			switch(binary_type)
			{ /* End switch binary_type */

				case BIOS:
					/* Set boot vector */
					/* Note - these are long words */

					writeRAM(0, 0x0bf00401, SIZE32, card, NORMAL);
					writeRAM(4, 0x0, SIZE32, card, NORMAL);
					writeRAM(0xc00, 0x0, SIZE16, card, NORMAL);

					break;

				case FEP:

					/* Clear confirm word */
					writeRAM(0xd20, 0x0, SIZE16, card, NORMAL);

					/* Set initial address and BIOS command */
					writeRAM(0xc34, 0xbfc01004L, SIZE32, card, NORMAL);
					writeRAM(0xc30, 0x3L, SIZE32, card, NORMAL);

					break;

			} /* End switch binary_type */

			/* Release reset now */
			outb(0x00, (int)di->port + 0);

			/* Wait for binary (BIOS or FEP) to boot */
			wait_for_binary(card, binary_type);

			break;
		} /* End case PCXEM */

		case PCXE:
		case PCXEVE:
		{ /* Begin case PCXE or PCXEVE */
			/* Put us back in window 0 */
			outb(0x80, (int)di->port + 1);

			switch(binary_type)
			{ /* End switch binary_type */

				case BIOS:

					/* Clear confirm word */
					writeRAM(0xc00, 0x0, SIZE16, card, NORMAL);

					/* Release reset now */
					outb(0x00, (int)di->port + 0);
	
					wait_for_binary(card, binary_type);

					break;

				case FEP:

					/*
						Do not forget to byte swap these for little endian.
						0c40 <- 0002h     BIOS request to move program
						0c42 <- f200      seg + 200h

                                    = f2000002h
						c44  <- 0000h
						c46  <- 0200h

						                  = 02000000h
						c48  <- 0000h
						c4a  <- 2000h	
						                  = 20000000h
					*/

					/* 
						Interestingly enough the 8E glasgo does not
						need a command to move the fep program since
						dual port ram is shadowed over local memory.

						Additionally while the Glasgo offers the standard
						8 * 8K windows, it also offers 2 * 32K windows.
						It does NOT support the older 64K window.

						The below command is not needed, but we
						will keep it for XEVE compatability.  As long
						as we use the 8 * 8K windows the XEVE and XE will
						look identical to the driver.  They only differ when 
						using the 2 * 32K Glasgo window mode or the 1 * 64K
						XEVE window mode.
					*/
 

					/* Form BIOS request to move program to 0x2000 */
					writeRAM(0xc40, 0xf2000002L, SIZE32, card, NORMAL);
					writeRAM(0xc44, 0x02000000L, SIZE32, card, NORMAL);
					writeRAM(0xc48, 0x20000000L, SIZE32, card, NORMAL);

					/* 
						NOTE : The int below is NOT asserted on 
						the host bus unless the function code in 0xc40
						= 0.  In this case ints are generated on the card 
						AND host.  In all other situations the int is 
						generated just on the card.
					*/
 
					/*
						Toggle board interrupt.  This is done to make the 
						BIOS run the command.
					*/
					outb(0x8, (int)di->port);
					outb(0x00, (int)di->port);


					for (index = 0; index < 4; index++)
					{
						if ((readRAM(0xc40, SIZE16, card, NORMAL)) == 0)
							break; /* We got a 0, we can leave this loop now ! */
						sleep(1);
					}

					if (index == 4)
						fatal("Attempt to toggle interrupt failed.\n");
						

					/* Form BIOS request to execute FEPOS */
					writeRAM(0xc40, 0x02000001L, SIZE32, card, NORMAL);
					writeRAM(0xc44, 0x4, SIZE16, card, NORMAL);

					/* Clear confirm word */
					writeRAM(0xd20, 0x0, SIZE16, card, NORMAL);

					/*
						Toggle board interrupt.  This is done to make the 
						BIOS run the command.
					*/
					outb(0x08, (int)di->port);
					outb(0x0, (int)di->port);

					/* Wait to see if the FEP comes up (BOOTS) */
					wait_for_binary(card, binary_type);



					/* Get command in pointer */
					cmdInPtr = readRAM(0xd10, SIZE16, card, NORMAL);

					/* Set buffer space command */
					writeRAM((0x400 + cmdInPtr), 0x92, SIZE8, card, NORMAL);
					writeRAM((0x401 + cmdInPtr), 0x0, SIZE8, card, NORMAL);

					/*
					 	Set up for 32L buffers allow each receive, transmit
						buffer to have its on 8K window.
					*/

					writeRAM((0x402 + cmdInPtr), 0x8000, SIZE16, card, NORMAL);

					/* Advance the command in pointer */
					writeRAM(0xd10, (cmdInPtr + 4), SIZE16, card, NORMAL);
					break;

			} /* End switch binary_type */

			break;
		} /* End case for PCXE and PCXEVE */

		case PCXI:
		case PC64XE: /* Works same as PCXI */
		{ /* Begin case PCXI and PC64XE */
			switch (binary_type)
			{ /* Begin  switch binary_type */

				case BIOS:

					/* Clear confirm word */
					writeRAM(0xc00, 0x0, SIZE16, card, NORMAL);

					/* Release reset now, but leave memory enabled */
					outb(0x02, (int)di->port);
	
					wait_for_binary(card, binary_type);

				break;

				case FEP:
					/*
						Do not forget to byte swap these for little endian.
						0c40 <- 0002h     BIOS request to move program
						0c42 <- f200      seg + 200h  (f000 for a 64K board) 

                                    = f2000002h
						c44  <- 0000h
						c46  <- 0200h

						                  = 02000000h
						c48  <- 0000h
						c4a  <- 2000h	
						                  = 20000000h
					*/

					outb(0x02, (int)di->port); /* Leave memory enabled */


					/* Form BIOS request to move program to 0x2000 */

					writeRAM(0xc40, (0x02000002L + (memory_seg << 16)),
					         SIZE32, card, NORMAL);
					writeRAM(0xc44, 0x02000000L, SIZE32, card, NORMAL);
					writeRAM(0xc48, 0x20000000L, SIZE32, card, NORMAL);
					/* 
						NOTE : The int below is NOT asserted on 
						the host bus unless the function code in 0xc40
						= 0.  In this case ints are generated on the card 
						AND host.  In all other situations the int is 
						generated just on the card.
					*/
 
					/*
						Toggle board interrupt.  This is done to make the 
						BIOS run the command.
					*/
					outb(0x0A, (int)di->port);
					outb(0x02, (int)di->port); /* Leave memory enabled */

					for (index = 0; index < 4; index++)
					{
						if ((readRAM(0xc40, SIZE16, card, NORMAL)) == 0)
							break; /* We got a 0, we can leave this loop now ! */
						sleep(1);
					}

					if (index == 4)
						fatal("Attempt to toggle interrupt failed.\n");
						

					/* Form BIOS request to execute FEPOS */
					writeRAM(0xc40, 0x02000001L, SIZE32, card, NORMAL);
					writeRAM(0xc44, 0x0004, SIZE16, card, NORMAL);

					/* Clear confirm word */
					writeRAM(0xd20, 0x0000, SIZE16, card, NORMAL);

					/*
						Toggle board interrupt.  This is done to make the 
						BIOS run the command.
					*/
					outb(0x0A, (int)di->port);
					outb(0x02, (int)di->port);
					/* Wait to see if the FEP comes up (BOOTS) */
					wait_for_binary(card, binary_type);
					break;

			} /* End switch binary_type */

			break;
		} /* End case PCXI and PC64XE */

	} /* End switch di->type */

	return(ret_val);	      

} /* End load_binary */

/* ------------------- Begin main ------------------------- */
void main(int argc, char *argv[])
{ /* Begin main */

	int digiFD, card;

	if (argc > 2)
	{
		printf("<Error> - Too many arguments\n");
		printf("digiDload [v]\n");
		exit(-1);
	}

	if (argc == 2)
	{
		/* strcmp returns false if expression is equal */
		if (!strcmp("v",argv[1]))
			verbose = TRUE;
		else
		{
			printf("<Error> - Invalid argument\n");
			printf("digiDload [v]\n");
			exit(-1);
		}

	}

	if ((digiFD = open("/dev/digi_ctl", O_RDWR)) < 0 )
	{
		perror(" cant open digi ctl ");
	}

	init(digiFD);
	for (card = 0; card < board_count; card++)
	{
		reset_cards(card);
		load_binary(card, BIOS);
		sleep(1); 
		load_binary(card, FEP);  
	}


	if (ioctl(digiFD, DIGI_INIT, 0) < 0  )
	{
		fatal("Can't ioctl DIGI_INIT %d\n",  errno); 
	}

	if (ioctl(digiFD, DIGI_POLLER, 0) < 0  )
	{
		fatal("Can't ioctl DIGI_POLLER %d\n", errno); 
	}

	printf("digiDload complete : Card initialized\n");

} /* End main */
