
/*
 *	Copyright (c) 1989-94 DigiBoard Incorporated.
 *	All Rights Reserved.
 */

#ifndef	NOT_STARTC_ASTOPC
#	define	STARTC_STOPC	1
#endif

#ifndef	NOT_ASTARTC_ASTOPC
#	define	ASTARTC_ASTOPC	1
#endif


#include <sys/types.h>
#include <sys/wait.h>
#include <sys/termio.h>
#include <sys/errno.h>
#include <sys/ioctl.h>

#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <signal.h>
#include <unistd.h>


#include "digi1.h"

extern int errno;

extern char *tgetstr();
extern char *strncpy();

#define	outdev	stdout

/* ----------------- Begin prototypes ------------------------ */
int digiset(int, digi_t *);
int digiget(int, digi_t *);
void ditty(int); 
void catch(void);

long	sbaud[16] = {
	0, 50, 75, 110, 134, 150, 200, 300, 600,
	1200, 1800, 2400, 4800, 9600, 19200, 38400
};

long fbaud[5] = {0, 57600, 115200, 230400, 460800};


#define STTY	"/bin/stty"
#define STTYA	"/bin/stty -a"

/*
 *	Command parameters.
 */
digi_t	digi;		/* digi parameters to set.		*/
int	flag0;		/* digi_flags to clear			*/
int	flag1;		/* digi_flags to set			*/
int	modem0;		/* Modem control bits to clear		*/
int	modem1;		/* Modem control bits to set		*/
int	flush[3];	/* Flush tty modes			*/
int	sendbreak;	/* Send a BREAK	flag			*/
int	tcxonc[4];	/* Stop/restart output/input modes	*/

#ifdef	STARTC_STOPC
unchar	startc;		/* Start flow control character		*/
unchar	stopc;		/* Stop flow control character		*/
#endif

#ifdef	ASTARTC_ASTOPC
unchar	astartc;	/* Aux. start flow control character	*/
unchar	astopc;		/* Aux. stop flow control character	*/
#endif


#define MAXPARM	100		/* Max STTY parameters		*/
#define MAXTTY	255		/* Max ttys to set up		*/

int	nparm;			/* Number of parameters		*/
char	*parm[MAXPARM+2];	/* List of parameters to set	*/

int	ntty;			/* Number of ttys		*/
char	*tty[MAXTTY];		/* List of ttys to set		*/

int	print;			/* Print the current (final)	*/
				/* tty settings			*/

/********************************************************
 *	digiset - Set Digiboard special parameters	*
 ********************************************************/

int digiset(int fd, digi_t *digi)
{
	return(ioctl(fd, DIGI_SETAW, digi)) ;
}



/********************************************************
 *	digiget - Get DigiBoard special parameters.	*
 ********************************************************/

int digiget(int fd, digi_t *digi)
{
	return(ioctl(fd, DIGI_GETA, digi)) ;
}

	

/********************************************************
 *	ditty - Perform ditty operations.		*
 ********************************************************/

void ditty(int fd) /* File descriptor of tty */
{ /* Begin ditty */

	int modem, eol, baud;
	digi_t di;
	digiflow_t	dflow;
	struct termio tbuf;

	/*
	 *	Send a break signal.
	 */
	if (sendbreak)
	{
		if (ioctl(fd, TCSBRK, 0) == -1)
		{
			perror("TCSBRK") ;
			return ;
		}
	}

	/*
	 *	Set Digiboard parameters.
	 */
	if	(flag0 | flag1)
	{ /* Begin if flag0 or flag1 set */

	
		if (digiget(fd, &di) == -1) 
		{
			perror("DIGI_GETA") ;
			return ;
		}

		di.digi_flags &= ~flag0 ;
		di.digi_flags |= flag1 ;


		if (digiset(fd, &di) == -1)
		{
			perror("ditty") ;
			return ;
		}
	} /* End if flag0 or flag1 set */

	/*
	 *	Set modem control lines.
	 */

	if (modem0)
	{
		if (ioctl(fd, TIOCMBIC, &modem0) == -1)
		{
			perror("TIOCMBIC") ;
			return ;
		}
	}

	if (modem1 & TIOCM_RTS) 
	{
		if (digiget(fd, &di) == -1)
		{
			perror("DIGI_GETA") ;
			return ;
		}

		if (di.digi_flags & RTSPACE)
			modem1 &= ~TIOCM_RTS;
	}

	if (modem1)
	{
		if (ioctl(fd, TIOCMBIS, &modem1) == -1)
		{
			perror("TIOCMBIS") ;
			return ;
		}
	}

#ifdef	STARTC_STOPC
	if (startc || stopc)
	{
		digiflow_t	digiflow;

		if (ioctl(fd, DIGI_GETFLOW, &digiflow) == -1)
		{
			perror("DIGI_GETFLOW");
			return ;
		}

		if ( startc )
			digiflow.startc = startc;

		if ( stopc )
			digiflow.stopc = stopc;

		if (ioctl(fd, DIGI_SETFLOW, &digiflow) == -1)
		{
			perror("DIGI_SETFLOW") ;
			return ;
		}
	}
#endif /* STARTC_STOPC */

#ifdef	ASTARTC_ASTOPC
	if (astartc || astopc)
	{
		digiflow_t	digiflow;

		if (ioctl(fd, DIGI_GETAFLOW, &digiflow) == -1)
		{
			perror("DIGI_GETAFLOW");
			return ;
		}

		if (astartc)
			digiflow.startc = astartc;

		if (astopc)
			digiflow.stopc = astopc;

		if (ioctl(fd, DIGI_SETAFLOW, &digiflow) == -1)
		{
			perror("DIGI_SETAFLOW") ;
			return ;
		}
	}
#endif /* ASTARTC_ASTOPC */

	/*
	 *	Set any stty parameters requested.
	 */

	if (nparm) 
	{ 
		int pid;

		parm[0] = "stty";

		while ((pid = fork()) == -1) sleep(1);

		if (pid == 0)
		{
			/* Do an stty */
			execvp(STTY, parm);
			/* If execvp works we wont get here */
			perror(STTY);
			exit(2);
		}

		while (wait((union wait *)0) != pid) ;
	}

	eol = 0 ;
	/*
	 *	Display Digiboard parameters.
	 */

	/* Find out what flags are on or off on Digi. */
	/* Covers IXON, FASTBAUD, FORCEDCD, ALTPIN, AIXON */
	if ((print && (digiget(fd, &di))) != -1)
	{
		/*
		 * used to acculate baud rates.
		 */
		
		/* ---------------------------------------------------------------
			It appears that the below call doesn't exist.  We need cflags 
			here, either that are a baud rate.  We need to either retrieve
			the cflags from the driver or baud from stty
		------------------------------------------------------------------ */
		if (ioctl(fd, TCGETA, &tbuf) == -1)
		{
			perror("TCGETA") ;
			return ;
		}

		/* The below returns an index into baud table RONNIE*/
		baud = tbuf.c_cflag & B38400;



		(void) fprintf(outdev, "%sforcedcd %saltpin %sfastbaud (%ld)",
			(di.digi_flags & DIGI_FORCEDCD) ? "" : "-",
			(di.digi_flags & DIGI_ALTPIN) ? "" : "-",
			(di.digi_flags & DIGI_FAST) ? "" : "-",
			(di.digi_flags & DIGI_FAST) ? fbaud[baud] : sbaud[baud]);
		(void) fprintf(outdev, "\n") ;


#ifndef	COMXI	/* comxi fep supports only rts and cts flow control.	*/

		(void) fprintf(outdev, "%srtspace %sdtrpace %sctspace %sdsrpace %sdcdpace",
			(di.digi_flags & RTSPACE) ? "" : "-",
			(di.digi_flags & DTRPACE) ? "" : "-",
			(di.digi_flags & CTSPACE) ? "" : "-",
			(di.digi_flags & DSRPACE) ? "" : "-",
			(di.digi_flags & DCDPACE) ? "" : "-");

#else

		(void) fprintf(outdev, "%sctspace %srtspace",
			(di.digi_flags & CTSPACE) ? "" : "-",
			(di.digi_flags & RTSPACE) ? "" : "-");

#endif /* COMXI */

		eol = 1;
	}

	if (eol) fprintf(outdev, "\n");


	eol = 0 ;
	/*
	 *	Display modem parameters.
	 */
	
	if (print && ioctl(fd, TIOCMGET, &modem) != -1)
	{
		if (eol) (void) fprintf(outdev, " ");

		(void) fprintf(outdev,
			"DTR%c  RTS%c  CTS%c  CD%c  DSR%c  RI%c",
			(modem & TIOCM_DTR) ? '+' : '-',
			(modem & TIOCM_RTS) ? '+' : '-',
			(modem & TIOCM_CTS) ? '+' : '-',
			(modem & TIOCM_CD) ? '+' : '-',
			(modem & TIOCM_DSR) ? '+' : '-',
			(modem & TIOCM_RI) ? '+' : '-') ;
		eol = 1 ;
	}

	if (eol) fprintf(outdev, "\n") ;

#ifdef	STARTC_STOPC
	if (print && ioctl(fd, DIGI_GETFLOW, &dflow) != -1)
	{
		(void) fprintf(outdev, "startc = 0x%x stopc = 0x%x\n",
			dflow.startc, dflow.stopc);
	}
#endif /* STARTC_STOPC */

#ifdef	ASTARTC_ASTOPC
	if (print && ioctl(fd, DIGI_GETAFLOW, &dflow) != -1)
	{
		(void) fprintf(outdev, "%saixon astartc = 0x%x astopc = 0x%x\n",
			(di.digi_flags & DIGI_AIXON) ? "" : "-",
			dflow.startc, dflow.stopc);
	}
#endif /* ASTARTC_ASTOPC */

	/*
	 *	Display stty parameters.
	 */

	fflush(outdev);

	if (print == 1)
	{
		(void) system(STTY) ;
	}

	if (print == 2)
	{
		(void) system(STTYA) ;
	}

} /* End ditty */

/********************************************************
 *    catch - Stub to catch interrupts                  *
 ********************************************************/

void catch(void) {}

/********************************************************
 *     Main program to process parameters.              *
 ********************************************************/

int main(int argc, char **argv)
{ /* Begin Main */

	int i, fd, newfd;
	int ac = argc;
	char ttname[20], *ap, *tp;

	/*
	 *	Process parameters.
	 */

	for (i = 1 ; i < argc ; i++)
	{ /* Begin for each arg */

		ap = argv[i] ;


#ifdef	STARTC_STOPC
		if (strcmp(ap, "startc") == 0 && i+1 < argc)
		{
			startc = (unchar) strtol(argv[++i], (char**)0, 0);
		}
		else 
		if (strcmp(ap, "stopc") == 0 && i+1 < argc)
		{
			stopc = (unchar) strtol(argv[++i], (char**)0, 0);
		}
#endif

#ifdef	ASTARTC_ASTOPC
		else 
		if (strcmp(ap, "astartc") == 0 && i+1 < argc)
		{
			astartc = (unchar) strtol(argv[++i], (char**)0, 0);
		}
		else 
		if (strcmp(ap, "astopc") == 0 && i+1 < argc)
		{
			astopc = (unchar) strtol(argv[++i], (char**)0, 0);
		}
#endif

		else 
		if (strcmp(ap, "-dtr") == 0)
		{
			modem0 |= TIOCM_DTR ;
		}
		else 
		if (strcmp(ap, "dtr") == 0)
		{
			modem1 |= TIOCM_DTR ;
		}
		else 
		if (strcmp(ap, "-rts") == 0)
		{
			modem0 |= TIOCM_RTS ;
		}
		else 
		if (strcmp(ap, "rts") == 0)
		{
			modem1 |= TIOCM_RTS ;
		}
		else 
		if (strcmp(ap, "rtspace") == 0)
		{
			flag1 |= RTSPACE;
		}
		else 
		if (strcmp(ap, "-rtspace") == 0)
		{
			flag0 |= RTSPACE;
		}
		else 
		if (strcmp(ap, "ctspace") == 0)
		{
			flag1 |= CTSPACE;
		}
		else 
		if (strcmp(ap, "-ctspace") == 0)
		{
			flag0 |= CTSPACE;
		}
		else 
		if (strcmp(ap, "dtrpace") == 0)
		{
			flag1 |= DTRPACE;
		}
		else 
		if (strcmp(ap, "-dtrpace") == 0)
		{
			flag0 |= DTRPACE;
		}
		else
		if (strcmp(ap, "dsrpace") == 0)
		{
			flag1 |= DSRPACE;
		}
		else 
		if (strcmp(ap, "-dsrpace") == 0)
		{
			flag0 |= DSRPACE;
		}
		else 
		if (strcmp(ap, "dcdpace") == 0)
		{
			flag1 |= DCDPACE;
		}
		else 
		if (strcmp(ap, "-dcdpace") == 0)
		{
			flag0 |= DCDPACE;
		}
		else 
		if (strcmp(ap, "flushin") == 0)
		{
			flush[0]++;
		}
		else 
		if (strcmp(ap, "flushout") == 0)
		{
			flush[1]++;
		}
		else 
		if (strcmp(ap, "flush") == 0)
		{
			flush[2]++;
		}
		else 
		if (strcmp(ap, "break") == 0)
		{
			sendbreak++ ;
		}
		else 
		if (strcmp(ap, "stopout") == 0)
		{
			tcxonc[0]++ ;
		}
		else 
		if (strcmp(ap, "startout") == 0)
		{
			tcxonc[1]++ ;
		}
		else 
		if (strcmp(ap, "stopin") == 0)
		{
			tcxonc[2]++ ;
		}
		else 
		if (strcmp(ap, "startin") == 0)
		{
			tcxonc[3]++ ;
		}
		else 
		if (strcmp(ap, "fastbaud") == 0)
		{
			flag1 |= DIGI_FAST;
		}
		else 
		if (strcmp(ap, "-fastbaud") == 0)
		{
			flag0 |= DIGI_FAST;
		}
		else 
		if (strcmp(ap, "forcedcd") == 0)
		{
			flag1 |= DIGI_FORCEDCD;
		}
		else 
		if (strcmp(ap, "-forcedcd") == 0)
		{
			flag0 |= DIGI_FORCEDCD;
		}
#ifdef	ASTARTC_ASTOPC
		else 
		if (strcmp(ap, "aixon") == 0)
		{
			flag1 |= DIGI_AIXON;
		}
		else 
		if (strcmp(ap, "-aixon") == 0)
		{
			flag0 |= DIGI_AIXON;
		}
#endif
		else 
		if (strcmp(ap, "altpin") == 0)
		{
			flag1 |= DIGI_ALTPIN;
		}
		else 
		if (strcmp(ap, "-altpin") == 0)
		{
			flag0 |= DIGI_ALTPIN;
		}
#ifdef PARA
		else 
		if (strcmp(ap, "-p") == 0 || strcmp(ap, "-parallel") == 0)
		{
			parallel = 1;	/* interpret as parallel port */
		}
		else 
		if (strcmp(ap, "output") == 0)
		{
			parallel = 1;
			flag0 |= DIGI_PP_INPUT;
		}
		else 
		if (strcmp(ap, "input") == 0)
		{
			parallel = 1;
			flag1 |= DIGI_PP_INPUT;
		}
		else 
		if (strcmp(ap, "noack") == 0)
		{
			parallel = 1;
			flag1 |= DSRPACE;
		}
		else 
		if (strcmp(ap, "-noack") == 0)
		{
			parallel = 1;
			flag0 |= DSRPACE;
		}
		else 
		if (strcmp(ap, "reset") == 0)
		{
			parallel = 1;
			preset = 1;
		}
		else 
		if (((strcmp(ap, "-OP") == 0 || strcmp(ap, "-op") == 0) ||
			  (strcmp(ap, "OP-") == 0 || strcmp(ap, "op-") == 0)))
		{
			parallel = 1;
			modem0 |= TIOCM_DTR ;
		}
		else 
		if (((strcmp(ap,  "OP") == 0 || strcmp(ap,  "op") == 0) ||
			  (strcmp(ap, "+OP") == 0 || strcmp(ap, "+op") == 0) ||
			  (strcmp(ap, "OP+") == 0 || strcmp(ap, "op+") == 0)))
		{
			parallel = 1;
			modem1 |= TIOCM_DTR ;
		}
		else 
		if ((strcmp(ap, "0.8") == 0 || strcmp(ap, ".8") == 0))
		{
			parallel = 1;
			pwidth = 13;
		}
		else 
		if ((strcmp(ap, "1.0") == 0 || strcmp(ap, "1") == 0))
		{
			parallel = 1;
			pwidth = 12;
		}
		else 
		if (strcmp(ap, "1.2") == 0)
		{
			parallel = 1;
			pwidth = 11;
		}
		else 
		if (strcmp(ap, "1.6") == 0)
		{
			parallel = 1;
			pwidth = 10;
		}
		else 
		if ((strcmp(ap, "2.0") == 0 || strcmp(ap, "2") == 0))
		{
			parallel = 1;
			pwidth = 9;
		}
		else 
		if ((strcmp(ap, "3.0") == 0 || strcmp(ap, "3") == 0))
		{
			parallel = 1;
			pwidth = 8;
		}
		else 
		if ((strcmp(ap, "4.0") == 0 || strcmp(ap, "4") == 0))
		{
			parallel = 1;
			pwidth = 7;
		}
		else 
		if ((strcmp(ap, "5.0") == 0 || strcmp(ap, "5") == 0))
		{
			parallel = 1;
			pwidth = 6;
		}
		else 
		if ((strcmp(ap, "10.0") == 0 || strcmp(ap, "10") == 0))
		{
			parallel = 1;
			pwidth = 5;
		}
		else 
		if ((strcmp(ap, "20.0") == 0 || strcmp(ap, "20") == 0))
		{
			parallel = 1;
			pwidth = 4;
		}
		else 
		if ((strcmp(ap, "30.0") == 0 || strcmp(ap, "30") == 0))
		{
			parallel = 1;
			pwidth = 3;
		}
		else 
		if ((strcmp(ap, "40.0") == 0 || strcmp(ap, "40") == 0))
		{
			parallel = 1;
			pwidth = 2;
		}
		else 
		if ((strcmp(ap, "50.0") == 0))
		{
			parallel = 1;
			pwidth = 1;
		}
#endif
		else 
		if (strcmp(ap, "-a") == 0)
		{
			print = 2 ;
		}
		else 
		if (strcmp(ap, "-n") == 0 && i+1 < argc) 
		{
			if (ntty >= MAXTTY)
			{
				(void) fprintf(stderr, 
				"%s: too many tty devices given!\n", argv[0]) ;
				exit(2) ;
			}
			tty[ntty++] = argv[++i];
			ac -= 2;
		} 
		else 
		if ((strncmp(ap, "/dev/", 5) == 0) || (strncmp(ap, "tty", 3) == 0) || (strncmp(ap, "cud", 3) == 0))
		{
			if (ntty >= MAXTTY)
			{
				(void) fprintf(stderr, 
					"%s: too many tty devices given!\n", argv[0]) ;
				exit(2) ;
			}

			tty[ntty++] = ap ;
			ac-- ;
		}
		else 
		{
			if (nparm >= MAXPARM)
			{
				(void) fprintf(stderr,
					"%s: too many stty parameters given!\n", argv[0]) ;
				exit(2) ;
			}
			parm[++nparm] = ap ;
		}

	} /* End for each arg */

	if (ac <= 1)
		print = 1;

	/*
	 *	If no tty was specified, set options on
	 *	standard output.
	 */

	if (!ntty)
	{
		ditty(0) ; /* standard input */
	}  
	else
	{ /* Begin tty's were specified */
		/*
		*	If a list of ttys were given, set options
		*	on all of them.
		*/

		(void) signal(SIGALRM, (void *)catch);

		for (i = 0 ; i < ntty ; i++)
		{ /* Begin for each tty */

			tp = tty[i] ;
			if (tp[0] != '/')
			{
				(void) sprintf(ttname, "/dev/%s", tp) ;
				tp = ttname ;
			}

			(void) alarm(1) ;
			fd = open(tp, O_WRONLY | O_NDELAY) ;
			(void) alarm(0) ;

			if (fd < 0)
			{
				if (errno == EINTR)
				{
					(void) fprintf(stderr,
						"%s: another process waiting on open\n", tp) ;
				}
				else perror(tp) ;
			}
			else
			{


				(void) close(0);
				/* This should always return file descriptor 0 (the lowest ) */
				newfd = dup(fd);

				/* 
					I don't really like the way this works.  By closing stdin
					we guarantee that the next 'dup'ed file descriptor will be
					0 (Because dup always uses the lowest available descriptor
					and stdin always has the lowest value of 0).  This is done
					so that the stty command 'exec'ed by ditty will be faked
					into believing that it is talking to standard input.  In this
					way he avoids sending arguments to the stty command.
				*/

				(void) close(fd);
				if (newfd < 0)
				{
					perror("dup failed ");
				}
				

				if (print && ntty > 1)
				{
					(void) fprintf(outdev, "\n####### %s\n", tp) ;
				}

				/* Should always be 0 */
				ditty(newfd) ;
			}
		} /* End for each tty */
	} /* End tty's were specified */

	return(0);

} /* End Main */

